/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

@import url("chrome://inspector/skin");

.textAttrsTextRange {
  border-bottom: medium dotted blue;
  padding: 7px;
  border-bottom-left-radius: 16px;
  border-bottom-right-radius: 16px;
  line-height: 250%;
  margin: 2px;
}

.textAttrsTextRange:hover {
  border-bottom: medium dotted red;
}

.textAttrsTextRange[selected] {
  border-bottom: medium groove purple;
}
