/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

@import url("chrome://inspector/skin");

treechildren::-moz-tree-checkbox(checked) {
  list-style-image: url("chrome://global/skin/checkbox/cbox-check.gif");
}

#welHandlerEditor {
  font-family: monospace;
}

#welGrippyButton {
  cursor: default;
}

#welSplitter {
  -moz-appearance: none;
}
