/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

dialog {
    padding-top: 0px;
    padding-left: 0px;
    padding-right: 0px;
    width: 30em;
}

.box-padded {
    background-color: Window;
    color: WindowText;
}

.large-text {
    font-size: large;
}

#logo {
    background: url(chrome://chatzilla/skin/images/logo.png);
    background-repeat: no-repeat;
    width:  32px;
    height: 32px;
}

#version {
    color: GrayText;
}

.contributors-label {
    margin-top: 0.5em;
    margin-bottom: 0.5em;
}

.contributors label {
    margin-top: 0px;
    margin-bottom: 0px;
}

#groove {
    margin-top: 0px;
}
