/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#rightPanel {
	width: 12em;
}

#bottomPanel {
	width: 50em;
	height: 20em;
}

.platform-Mac #rightPanel {
	width: 18em;
}

#loadContainer {
	margin: 0;
}

#loadBar {
	-moz-appearance: none;
}
